import { ArrayChange, DiffArrayResult, EqualFn, HDocOperation, Id, IDocumentSchema, INormalizedDocument, INormalizedMutableMapsDocument, IParentedId, MappedParentedTypesFields, MapsOfNormDoc, UOfNormDoc, WasTouchedFn } from './HTypes';
/**
 * Returns a list of HDocOperations that if applied
 * will transform baseDoc in laterDoc. laterDoc
 * is assumed to have the same root element and schema as
 * baseDoc
 *
 * @param {INormalizedDocument<MapsInterface, U>} baseDoc
 * @param {INormalizedDocument<MapsInterface, U>} laterDoc
 * @returns {HDocOperation<MapsInterface, any, U>[]}
 */
export declare function diff<NorDoc extends INormalizedDocument<any, any>>(baseDoc: NorDoc, laterDoc: NorDoc): HDocOperation<MapsOfNormDoc<NorDoc>, UOfNormDoc<NorDoc>, any>[];
/**
 * Given a base version and a later version of a document element
 * with schema schema and type elementType, returns a Partial version
 * of the element containing the info fields that have changed between baseEl
 * and laterEl.
 *
 * @param {IDocumentSchema<MapsInterface, U>} schema
 * @param {U} elementType
 * @param {T} baseEl
 * @param {T} laterEl
 * @returns {Partial<T>}
 */
export declare function diffElementInfo<MapsInterface, U extends keyof MapsInterface, T extends IParentedId<U, U>>(schema: IDocumentSchema<MapsInterface, U>, elementType: U, baseEl: T, laterEl: T): Partial<T>;
export declare function diffInfoOf<MapsInterface, U extends keyof MapsInterface, T extends MappedParentedTypesFields<MapsInterface>>(baseDoc: INormalizedDocument<MapsInterface, U> | INormalizedMutableMapsDocument<MapsInterface, U>, laterDoc: INormalizedDocument<MapsInterface, U> | INormalizedMutableMapsDocument<MapsInterface, U>, elementType: U, elementId: Id): Partial<T>;
export declare const defaultEquals: (a: any, b: any) => boolean;
export declare const defaultWasTouchedFn: (_: any) => boolean;
/**
 * Creates a diff between two arrays, returning a list of change operations
 * that would bring the base array to the later array.
 *
 * Elements are considered equal based on the equalsFn parameter, that by
 * default uses the === boolean operator. You can use equals from lodash for
 * deep compares, for instance.
 *
 * The diff result returns both the list of sequential changes in the changes
 * member, and an array that has the same number of elements as the base array.
 * Each element shows what happened to the base element in the later array:
 * kept in the same position, moved to the left of the array or deleted.
 *
 * @param {T[]} base
 * @param {T[]} later
 * @param {EqualFn} equalsFn
 * @returns {DiffArrayResult<T>}
 */
export declare function diffArray<T>(base: T[], later: T[], { equalsFn, wasTouchedFn }?: {
    equalsFn?: EqualFn;
    wasTouchedFn?: WasTouchedFn<T>;
}): DiffArrayResult<T>;
/**
 * Given a base array and an array of array changes, returns a shallow copied
 * version of the base array to which all the changes have been applied in
 * sequence.
 *
 * @param {T[]} base
 * @param {ArrayChange<T>[]} changes
 * @returns {T[]}
 */
export declare function applyArrayDiff<T>(base: T[], changes: ArrayChange<T>[]): T[];
/**
 * An Id function allows creating a uniquely identifying string for
 * an element.
 */
export interface IdFn<T> {
    (value: T): string | T;
}
/**
 * Least common sequence result. buffer1[buffer1index] will be equal to
 * buffer2[buffer2index], and chain points to the next character in the least common
 * sequence, if it exists.
 */
export interface LcsResult {
    buffer1index: number;
    buffer2index: number;
    chain: null | LcsResult;
}
export interface DiffProps<T> {
    equalsFn?: EqualFn;
    wasTouchedFn?: WasTouchedFn<T>;
}
interface LcsProps<T> {
    wasTouchedFn: WasTouchedFn<T>;
    idFn: IdFn<T>;
}
export declare function longestCommonSequence<T>(buffer1: T[], buffer2: T[], inpProps?: Partial<LcsProps<T>>): LcsResult;
interface Diff3Options<T = any> {
    wasTouchedFn: (val: T, side: 'left' | 'right') => boolean;
}
export interface StableRegion<T> {
    stable: true;
    buffer: 'a' | 'o' | 'b';
    bufferStart: number;
    bufferLength: number;
    bufferContent: T[];
}
export interface UnstableRegion<T> {
    stable: false;
    aStart: number;
    aLength: number;
    aContent: T[];
    bStart: number;
    bLength: number;
    bContent: T[];
    oStart: number;
    oLength: number;
    oContent: T[];
}
export type DiffMergeRegion<T> = StableRegion<T> | UnstableRegion<T>;
export declare function diff3MergeRegions<T>(a: T[], o: T[], b: T[], inpProps?: Partial<Diff3Options>): DiffMergeRegion<T>[];
export interface OkMergeRegion<T> {
    ok: T[];
}
type ArrayElement<T> = T extends Array<infer S> ? S : T;
export interface ConflictMergeRegion<T> {
    conflict: {
        a: T[];
        aIndex: number;
        b: T[];
        bIndex: number;
        o: T[];
        oIndex: number;
    };
}
export type MergeRegion<T> = OkMergeRegion<T> | ConflictMergeRegion<T>;
export interface Diff3MergeOptions<T = any> extends Diff3Options<T> {
    excludeFalseConflicts: boolean;
    stringSeparator: string | RegExp;
}
export declare function diff3Merge<T extends string | Array<any>>(inpA: T, inpO: T, inpB: T, inpOptions?: Partial<Diff3MergeOptions>): MergeRegion<ArrayElement<T>>[];
export {};
