"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addElementToArrayReducer = void 0;
exports.clearedNormalizedDocument = clearedNormalizedDocument;
exports.cloneNormalizedDocument = cloneNormalizedDocument;
exports.docReducer = void 0;
exports.idAndTypeForPath = idAndTypeForPath;
exports.idAndTypeOfChange = idAndTypeOfChange;
exports.mutableDocument = mutableDocument;
exports.parentToChildFieldName = parentToChildFieldName;
exports.pathForElementWithId = pathForElementWithId;
exports.removeElementFromArrayReducer = void 0;
var _lodash = require("lodash");
var _LazyMap = require("./LazyMap");
var _HTypes = require("./HTypes");
var _HUtils = require("./HUtils");
var _HVisit = require("./HVisit");
function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); enumerableOnly && (symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; })), keys.push.apply(keys, symbols); } return keys; }
function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = null != arguments[i] ? arguments[i] : {}; i % 2 ? ownKeys(Object(source), !0).forEach(function (key) { _defineProperty(target, key, source[key]); }) : Object.getOwnPropertyDescriptors ? Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)) : ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } return target; }
function _defineProperty(obj, key, value) { key = _toPropertyKey(key); if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }
function _toPropertyKey(arg) { var key = _toPrimitive(arg, "string"); return typeof key === "symbol" ? key : String(key); }
function _toPrimitive(input, hint) { if (typeof input !== "object" || input === null) return input; var prim = input[Symbol.toPrimitive]; if (prim !== undefined) { var res = prim.call(input, hint || "default"); if (typeof res !== "object") return res; throw new TypeError("@@toPrimitive must return a primitive value."); } return (hint === "string" ? String : Number)(input); }
/**
 * Creates a shallow clone of the document. The maps are new objects,
 * but the entities mapped are the same as the original.
 *
 * The rationale for the shallow version of the elements is that
 * changes will be performed as setting new versions in the dictionary, rather
 * than direct manipulation of the objects.
 *
 * @param {INormalizedDocument<MapsInterface, U>} doc The document to be cloned
 * @returns {INormalizedDocument<MapsInterface, U>} shallow clone of the document
 */
function cloneNormalizedDocument(doc) {
  const clonedMaps = {};
  let mapField;
  for (mapField in doc.maps) {
    const entityMap = doc.maps[mapField];
    if ((0, _HUtils.isParentedMap)(entityMap)) {
      // eslint-disable-next-line @typescript-eslint/ban-ts-comment
      // @ts-ignore
      clonedMaps[mapField] = new Map(entityMap);
    }
  }
  return _objectSpread(_objectSpread({}, doc), {}, {
    maps: clonedMaps
  });
}

/**
 * Creates an empty copy version of the normalized document
 * passed as parameter.
 *
 * @returns {INormalizedDocument<MapsInterface, U>}
 */

function clearedNormalizedDocument(doc) {
  const clone = cloneNormalizedDocument(doc);
  let mapField;
  for (mapField in clone.maps) {
    const entityMap = clone.maps[mapField];
    if (!(0, _HUtils.isParentedMap)(entityMap)) continue;
    const parentedMap = entityMap;
    if (mapField === clone.rootType) {
      const rootsIdsToDelete = Array.from(parentedMap.keys()).filter(rootId => rootId !== clone.rootId);
      for (const rootIdToDelete of rootsIdsToDelete) {
        parentedMap.delete(rootIdToDelete);
      }
      const rootElement = parentedMap.get(clone.rootId);
      if (!rootElement) {
        throw new TypeError('Invalid document');
      }
      const rootFieldsMap = doc.schema.types[clone.rootType];
      for (const rootField in rootFieldsMap) {
        if (rootField === 'parentId') continue;
        if (Array.isArray(rootFieldsMap[rootField])) {
          const rootFieldVal = rootElement[rootField];
          if (!Array.isArray(rootFieldVal)) continue;
          rootElement[rootField] = [];
        } else {
          const rootFieldVal = rootElement[rootField];
          if (rootFieldVal !== null && rootFieldVal !== undefined) {
            rootElement[rootField] = null;
          }
        }
      }
    } else {
      parentedMap.clear();
    }
  }
  return clone;
}

/**
 * A Type link links an Element Type A to another element type U, where
 * element of types Y are either children or parent of the element of type A.
 * */

const schemaMap = new Map();
function getSchemaTypeMap(schema) {
  if (!schemaMap.has(schema.documentType)) {
    schemaMap.set(schema.documentType, createTypesMapForSchema(schema));
  }
  return schemaMap.get(schema.documentType);
}
function idAndTypeForPath(doc, path) {
  const {
    schema
  } = doc;
  if (!(path && Array.isArray(path) && path.length > 0)) {
    return {
      __typename: doc.rootType,
      _id: doc.rootId
    };
  }
  let trimmedPath = path.slice();
  let docContext = doc.maps[doc.rootType].get(doc.rootId);
  let schemaContext = schema.types[schema.rootType];
  for (let i = 0; trimmedPath.length > 0 && i < 10000; i++) {
    const {
      remainingPath,
      element
    } = nextElementTypeAndId(doc, docContext, schemaContext, trimmedPath);
    const {
      _id,
      __typename
    } = element;
    trimmedPath = remainingPath;
    if (trimmedPath.length > 0) {
      docContext = (0, _HUtils.mappedElement)(doc.maps, __typename, _id);
      if (!(__typename in schema.types)) {
        throw new Error(`TYpe not found in schema: ${String(__typename)}`);
      }
      schemaContext = schema.types[__typename];
    } else {
      return {
        __typename,
        _id
      };
    }
  }
  throw new Error('Entity not found for path');
}

/**
 * Function that navigates the hierarchy of the document and of the schema to retrieve
 * the next record type and id based on the subPath passed as parameter.
 *
 * @param {OMutableDocument<NormalizedDoc, U>} doc the document we are
 * navigating
 * @param {Id | IParentedId | OMutableDocument<NormalizedDoc, U>} docContext the current
 * subtree of the document we are working on.
 * @param {EntityReferences<U> | EntityReference<U>} schemaContext the current schema
 * subtree we are working on
 * @param {Path} subPath the path we are looking at to navigate. It is based on the
 * current schema and document contexts
 * @returns {{element: {type: U; _id: Id}; remainingPath: Path}}
 */
function nextElementTypeAndId(doc, docContext, schemaContext, subPath) {
  const typeMap = getSchemaTypeMap(doc.schema);
  const remainingPath = subPath.slice();
  if ((0, _HUtils.isSchemaReference)(schemaContext)) {
    const {
      __schemaType
    } = schemaContext;
    if (!typeMap.has(__schemaType)) {
      throw new Error(`Schema does not contain type ${__schemaType}`);
    }
    if (!((0, _HUtils.isIParentedId)(docContext) || (0, _HUtils.isId)(docContext))) {
      throw new Error(`Document error: expecting to be an entity`);
    }
    return {
      element: {
        __typename: __schemaType,
        _id: (0, _HUtils.isIParentedId)(docContext) ? docContext._id : docContext
      },
      remainingPath
    };
  } else if (Array.isArray(schemaContext)) {
    if (!(remainingPath.length >= 1 && typeof remainingPath[0] === 'number' && (0, _HUtils.isSchemaReference)(schemaContext[0]) && Array.isArray(docContext) && docContext.length > remainingPath[0])) {
      throw new Error('Expected an index for an array in the path');
    }
    const index = remainingPath[0];
    remainingPath.shift();
    return {
      element: {
        __typename: schemaContext[0].__schemaType,
        _id: docContext[index]
      },
      remainingPath
    };
  } else if (typeof schemaContext === 'object') {
    if (!(remainingPath.length >= 1 && typeof remainingPath[0] === 'string' && remainingPath[0] in schemaContext && remainingPath[0] in docContext)) {
      throw new Error('Expecting a field within the root document and schema');
    }
    const nextSchemaContext = schemaContext[remainingPath[0]];
    const nextDocContext = docContext[remainingPath[0]];
    remainingPath.shift();
    return nextElementTypeAndId(doc, nextDocContext, nextSchemaContext, remainingPath);
  }
  throw new Error('Schema/document error');
}
function parentTypeOfElement(document, childType, childId) {
  const typeMap = getSchemaTypeMap(document.schema);
  const childElement = (0, _HUtils.mappedElement)(document.maps, childType, childId);
  let parentType = null;
  if (childElement.parentType) {
    parentType = childElement.parentType;
  }
  if (!parentType) {
    const elementMappings = typeMap.get(childType);
    if (!(elementMappings && elementMappings.parent)) {
      throw new TypeError(`Type mappings not found for type ${String(childType)}`);
    }
    parentType = elementMappings.parent.type;
  }
  return parentType;
}
function parentToChildTypeMappings(document, parentType, childType) {
  const typeMap = getSchemaTypeMap(document.schema);
  const parentMappings = typeMap.get(parentType);
  if (!parentMappings) {
    throw new TypeError(`Parent type mappings for child type ${String(childType)}`);
  }
  if (!(childType in parentMappings.children)) {
    throw new TypeError(`Child type ${String(childType)} not found among children types of parent type ${String(parentType)}`);
  }
  return parentMappings.children[childType];
}
function parentToChildFieldName(document, parentType, childType) {
  const linkedFieldProps = parentToChildTypeMappings(document, parentType, childType);
  return linkedFieldProps.field;
}

/**
 * Creates a dictionary of types for the schema passed as parameter.
 *
 * The dictionary is used to work out the parent / child relation when
 * navigating a normalized documents from a node up to the root of the document.
 *
 * @param {IDocumentSchema<NormalizedDoc, U>} schema
 * @returns {TypeLinkDictionary<NormalizedDocument, U>}
 */
function createTypesMapForSchema(schema) {
  const childMap = new Map();
  for (const typeName in schema.types) {
    const typeSettings = schema.types[typeName];
    for (const fieldName in typeSettings) {
      if (fieldName === 'parentId') continue;
      const fieldValue = typeSettings[fieldName];
      const childFieldSettings = fieldValue[0];
      const existingSettings = childMap.get(childFieldSettings.__schemaType) || {};
      childMap.set(childFieldSettings.__schemaType, _objectSpread(_objectSpread({}, existingSettings), {}, {
        [typeName]: fieldName
      }));
    }
  }
  const typeLinkMap = new Map();
  for (const typeName in schema.types) {
    const typeSettings = schema.types[typeName];
    const typeEntry = {
      children: {}
    };
    for (const fieldName in typeSettings) {
      if (fieldName === 'parentId') continue;
      const fieldSettingsValue = typeSettings[fieldName];
      const childFieldSettings = fieldSettingsValue[0];
      typeEntry.children[childFieldSettings.__schemaType] = {
        type: childFieldSettings.__schemaType,
        field: fieldName,
        isArray: Array.isArray(fieldSettingsValue)
      };
    }
    typeLinkMap.set(typeName, typeEntry);
  }
  return typeLinkMap;
}

/**
 * Given an element type and id of and element within the document,
 * returns the path to it.
 *
 * Each element is expected to be referenced only once as an official child
 * within the document.
 *
 * The element could still be referenced informally from fields not captured
 * in the document schema
 *
 * @param {U} elementTypeMap
 * @param {Id} elementId
 * @returns {Path}
 */

function pathForElementWithId(doc, elementTypeMap, elementId) {
  if (!(elementTypeMap in doc.maps && typeof doc.maps[elementTypeMap] === 'object')) {
    throw new Error(`Map for type ${elementTypeMap} not found`);
  }
  const map = doc.maps[elementTypeMap];
  const element = map.get(elementId);
  if (!element) {
    throw new Error(`Referential integrity error with type ${elementTypeMap} and id ${elementId}`);
  }
  const path = [];
  if (element.parentId) {
    const parentType = parentTypeOfElement(doc, elementTypeMap, elementId);
    const parentToUsFieldLink = parentToChildTypeMappings(doc, parentType, elementTypeMap);
    // We are not a root element, let's go up the hierarchy
    const parentElement = (0, _HUtils.mappedElement)(doc.maps, parentType, element.parentId);
    if (!parentElement) {
      throw new Error(`Referential integrity error for parent of element ${elementTypeMap}.${elementId}`);
    }
    path.push(parentToUsFieldLink.field);
    if (parentToUsFieldLink.isArray) {
      const position = parentElement[parentToUsFieldLink.field].findIndex(parentedElementId => parentedElementId === elementId);
      if (position === -1) {
        throw new Error(`Element ${elementTypeMap}.${elementId} not found in parent`);
      }
      path.push(position);
    }
    const parentPath = pathForElementWithId(doc, parentType, element.parentId);
    return parentPath.concat(path);
  } else {
    if (!(elementTypeMap === doc.rootType && elementId === doc.rootId)) {
      throw new Error(`Top level element of type ${elementTypeMap} is not the root element of the document`);
    }
  }
  return path;
}
function mutableDocument(doc) {
  const schema = doc.schema;
  const typeMap = getSchemaTypeMap(schema);

  /**
   * Helper function that removes the reference to an element via
   * its id from its parent. Used in MutableDocument when moving or
   * deleting an element.
   *
   * Looks up in the schema to what parent type to look out for,
   * checking that the child is indeed referenced before removing it.
   *
   * @param {IMutableDocument<MapsInterface, U>} doc
   * @param {U} childType
   * @param {Id} childId
   * @private
   */
  function _removeElementFromParentContext(doc, childType, childId) {
    const toRemoveElement = (0, _HUtils.mappedElement)(doc.maps, childType, childId);
    if (toRemoveElement.parentId) {
      const parentId = toRemoveElement.parentId;
      const parentType = parentTypeOfElement(doc, childType, childId);
      const {
        field: parentToElementFieldName
      } = parentToChildTypeMappings(doc, parentType, childType);
      const parentElement = (0, _HUtils.mappedElement)(doc.maps, parentType, parentId);
      const parentChildField = parentElement[parentToElementFieldName];
      if (parentChildField && Array.isArray(parentChildField)) {
        const elementIndex = parentChildField.indexOf(toRemoveElement._id);
        if (elementIndex === -1) {
          throw new ReferenceError('Child element to delete not found in parent element');
        }
        const updatedList = parentChildField.slice();
        updatedList.splice(elementIndex, 1);
        doc.maps[parentType].set(parentElement._id, _objectSpread(_objectSpread({}, parentElement), {}, {
          [parentToElementFieldName]: updatedList
        }));
      } else {
        if (!(parentToElementFieldName in parentElement)) {
          throw new TypeError(`Field ${parentToElementFieldName} not found in parent object of type ${parentType}`);
        }
        doc.maps[parentType].set(parentElement._id, _objectSpread(_objectSpread({}, parentElement), {}, {
          [parentToElementFieldName]: null
        }));
      }
    }
  }

  /**
   * Helper function that inserts an element, via its ID, to
   * a parent element context.
   *
   * The function modifies both the parent, by referencing to the child element,
   * and the child element, by updating its parentId field.
   *
   * The function also ensures that the operation respects referential integrity
   * and schema integrity.
   *
   * @param {IMutableDocument<MapsInterface, U>} doc
   * @param {U} parentType
   * @param {Id} parentId
   * @param {SubEntityPathElement} positionInParent
   * @param {IParentedId} childElement
   * @returns {U}
   * @private
   */
  function _addElementIdToParentContext(doc, parentType, parentId, positionInParent, childElement) {
    const context = (0, _HUtils.mappedElement)(doc.maps, parentType, parentId);
    const parentSchema = doc.schema.types[parentType];
    if (!parentSchema) {
      throw new TypeError(`Schema for parent type ${parentType} not found`);
    }
    const field = typeof positionInParent === 'string' ? positionInParent : positionInParent[0];
    if (!(typeof context === 'object' && field in context && field in parentSchema)) {
      throw new Error('Position not found for insert operation');
    }
    const {
      __schemaType
    } = parentSchema[field][0];
    doc.maps[__schemaType].set(childElement._id, childElement.parentId === parentId ? childElement : _objectSpread(_objectSpread({}, childElement), {}, {
      parentId,
      parentType
    }));
    if (positionInParent instanceof Array && positionInParent.length === 2) {
      // expect string and index tuple
      const targetList = context[field].slice();
      const index = positionInParent[1] < 0 ? 0 : positionInParent[1] > targetList.length ? targetList.length : positionInParent[1];
      targetList.splice(index, 0, childElement._id);
      doc.maps[parentType].set(parentId, _objectSpread(_objectSpread({}, context), {}, {
        [field]: targetList
      }));
    } else {
      throw new Error('Unexpected position object');
    }
    return __schemaType;
  }
  let lazyMaps = {};
  for (const entityType in doc.maps) {
    lazyMaps = _objectSpread(_objectSpread({}, lazyMaps), {}, {
      [entityType]: new _LazyMap.LazyMutableMap(doc.maps[entityType])
    });
  }
  const mutableDoc = {
    maps: lazyMaps,
    schema,
    originalDocument: doc,
    rootType: doc.rootType,
    rootId: doc.rootId,
    changes: [],
    updatedDocument: function () {
      const updatedMaps = {};
      let hasChanges = false;
      for (const mapName in this.maps) {
        // @ts-expect-error
        updatedMaps[mapName] = this.maps[mapName].getMap();
        hasChanges = hasChanges || this.maps[mapName].hasChanged();
      }
      return hasChanges ? _objectSpread(_objectSpread({}, this.originalDocument), {}, {
        maps: updatedMaps
      }) : this.originalDocument;
    },
    applyChanges: function (changes) {
      const changesToRun = Array.isArray(changes) ? changes : [changes];
      for (const command of changesToRun) {
        if (command.__typename === _HTypes.HDocCommandType.INSERT_ELEMENT) {
          this.insertElement(command);
        }
        if (command.__typename === _HTypes.HDocCommandType.CHANGE_ELEMENT) {
          this.changeElement(command);
        }
        if (command.__typename === _HTypes.HDocCommandType.DELETE_ELEMENT) {
          this.deleteElement(command);
        }
        if (command.__typename === _HTypes.HDocCommandType.MOVE_ELEMENT) {
          this.moveElement(command);
        }
      }
    },
    insertElement: function (insertCmd) {
      const {
        element,
        parent,
        position
      } = insertCmd;
      if (!(0, _HUtils.isSubEntityPathElement)(position)) {
        throw new Error('Incorrect position parameter');
      }
      const {
        __typename: parentType,
        _id: parentId
      } = (0, _HUtils.isElementId)(parent) ? parent : this.idAndTypeForPath(parent);
      const elementId = element._id ? element._id : (0, _HUtils.generateNewId)();
      const newElement = _objectSpread(_objectSpread({}, element), {}, {
        _id: elementId
      });
      _addElementIdToParentContext(this, parentType, parentId, position, newElement);
      this.changes.push(_objectSpread(_objectSpread({}, insertCmd), {}, {
        element: _objectSpread(_objectSpread({}, element), {}, {
          _id: elementId
        }),
        parent: {
          __typename: parentType,
          _id: parentId
        }
      }));
      return newElement;
    },
    changeElement: function (changeCommand) {
      const {
        element,
        changes
      } = changeCommand;
      const {
        __typename,
        _id
      } = (0, _HUtils.isElementId)(element) ? element : this.idAndTypeForPath(element);
      const existingElement = (0, _HUtils.mappedElement)(this.maps, __typename, _id);
      const updatedElement = _objectSpread(_objectSpread({}, existingElement), changes);
      if ((0, _lodash.isEqual)(updatedElement, existingElement)) return;
      this.maps[__typename].set(updatedElement._id, updatedElement);
      this.changes.push((0, _HUtils.isElementId)(element) ? changeCommand : _objectSpread(_objectSpread({}, changeCommand), {}, {
        element: {
          __typename,
          _id
        }
      }));
    },
    moveElement: function (moveCommand) {
      const {
        changes,
        element,
        toParent,
        toPosition
      } = moveCommand;
      // 1. Find the element
      const {
        __typename,
        _id
      } = (0, _HUtils.isElementId)(element) ? element : this.idAndTypeForPath(element);
      const elementToMove = (0, _HUtils.mappedElement)(this.maps, __typename, _id);
      const elementSchemaContext = typeMap.get(__typename);
      if (!elementSchemaContext) {
        throw new TypeError(`Schema for type ${__typename} not found`);
      }

      // 2. Find the original parent - if any and remove reference to element
      const originalParentId = elementToMove.parentId || null;
      if (originalParentId) {
        _removeElementFromParentContext(this, __typename, _id);
      }

      // 3. Find the new parent
      const {
        __typename: targetParentType,
        _id: targetParentId
      } = (0, _HUtils.isElementId)(toParent) ? toParent : this.idAndTypeForPath(toParent);

      // 4. Add reference to element to new parent, while applying additional changes
      // if provided int he command
      _addElementIdToParentContext(this, targetParentType, targetParentId, toPosition, changes ? _objectSpread(_objectSpread({}, elementToMove), changes) : elementToMove);
      this.changes.push((0, _HUtils.isElementId)(moveCommand.element) && (0, _HUtils.isElementId)(moveCommand.toParent) ? moveCommand : _objectSpread(_objectSpread({}, moveCommand), {}, {
        element: {
          __typename,
          _id
        },
        toParent: {
          __typename: targetParentType,
          _id: targetParentId
        }
      }));
    },
    deleteElement: function (deleteCommand) {
      const {
        element
      } = deleteCommand;
      const {
        __typename,
        _id
      } = (0, _HUtils.isElementId)(element) ? element : this.idAndTypeForPath(element);
      const toDeleteElement = (0, _HUtils.mappedElement)(this.maps, __typename, _id);
      if (toDeleteElement.parentId) {
        _removeElementFromParentContext(this, __typename, _id);
      }
      const elementsToDelete = [];
      (0, _HVisit.visitDocument)(this, (_, nodeType, nodeId) => {
        elementsToDelete.push({
          __typename: nodeType,
          _id: nodeId
        });
      }, {
        traversal: _HTypes.DocumentVisitTraversal.DEPTH_FIRST,
        startElement: {
          type: __typename,
          _id
        }
      });
      for (const elementToDeleteId of elementsToDelete) {
        this.maps[elementToDeleteId.__typename].delete(elementToDeleteId._id);
      }
      this.changes.push((0, _HUtils.isElementId)(deleteCommand.element) ? deleteCommand : _objectSpread(_objectSpread({}, deleteCommand), {}, {
        element: {
          __typename,
          _id
        }
      }));
    },
    idAndTypeForPath: function (path) {
      return idAndTypeForPath(this, path);
    },
    pathForElementWithId: function (elementTypeMap, elementId) {
      if (!(elementTypeMap in this.maps && typeof this.maps[elementTypeMap] === 'object')) {
        throw new Error(`Map for type ${elementTypeMap} not found`);
      }
      const map = this.maps[elementTypeMap];
      const element = map.get(elementId);
      if (!element) {
        throw new Error(`Referential integrity error with type ${elementTypeMap} and id ${elementId}`);
      }
      const path = [];
      if (element.parentId) {
        const parentType = parentTypeOfElement(this, elementTypeMap, elementId);
        const parentToUsFieldLink = parentToChildTypeMappings(this, parentType, elementTypeMap);
        // We are not a root element, let's go up the hierarchy
        const parentElement = (0, _HUtils.mappedElement)(this.maps, parentType, element.parentId);
        if (!parentElement) {
          throw new Error(`Referential integrity error for parent of element ${elementTypeMap}.${elementId}`);
        }
        path.push(parentToUsFieldLink.field);
        if (parentToUsFieldLink.isArray) {
          const position = parentElement[parentToUsFieldLink.field].findIndex(parentedElementId => parentedElementId === elementId);
          if (position === -1) {
            throw new Error(`Element ${elementTypeMap}.${elementId} not found in parent`);
          }
          path.push(position);
        }
        const parentPath = this.pathForElementWithId(parentType, element.parentId);
        return parentPath.concat(path);
      } else {
        if (!(elementTypeMap === this.rootType && elementId === this.rootId)) {
          throw new Error(`Top level element of type ${elementTypeMap} is not the root element of the document`);
        }
      }
      return path;
    }
  };
  return mutableDoc;
}
const docReducer = (doc, cmd) => {
  const cmds = Array.isArray(cmd) ? cmd : [cmd];
  if (cmds.length < 1) return doc;
  const mutableDoc = mutableDocument(doc);
  try {
    mutableDoc.applyChanges(cmds);
  } catch (err) {
    // silently eaten
  }
  return mutableDoc.updatedDocument();
};
exports.docReducer = docReducer;
const removeElementFromArrayReducer = (doc, elementType, elementId, arrayFieldName, arrayElement) => {
  if (!(0, _HUtils.hasMappedElement)(doc, elementType, elementId)) return doc;
  const element = (0, _HUtils.mappedElement)(doc, elementType, elementId);
  if (arrayFieldName in element) {
    const elementArray = element[arrayFieldName];
    const index = elementArray.indexOf(arrayElement);
    if (index !== -1) {
      const updatedArray = elementArray.slice();
      updatedArray.splice(index, 1);
      const updateChange = {
        __typename: _HTypes.HDocCommandType.CHANGE_ELEMENT,
        element: {
          __typename: elementType,
          _id: elementId
        },
        // @ts-expect-error
        changes: {
          [arrayFieldName]: updatedArray
        }
      };
      return docReducer(doc, updateChange);
    }
  }
  return doc;
};

/**
 * Creates a new document with an array field updated with the
 * additional element(s) passed as parameters, optionally from a specific
 * index.
 *
 * @param {NorDoc} doc
 * @param {UOfNormDoc<NorDoc>} elementType
 * @param {Id} elementId
 * @param {AllMappedTypesFields<MapsOfNormDoc<NorDoc>>} arrayFieldName
 * @param {T} arrayElement
 * @param {number} insertIntoIndex
 * @returns {NorDoc}
 */
exports.removeElementFromArrayReducer = removeElementFromArrayReducer;
const addElementToArrayReducer = (doc, elementType, elementId, arrayFieldName, arrayElement, insertIntoIndex = -1) => {
  if (!(0, _HUtils.hasMappedElement)(doc, elementType, elementId)) return doc;
  const element = (0, _HUtils.mappedElement)(doc, elementType, elementId);
  if (arrayFieldName in element) {
    const elementArray = element[arrayFieldName];
    const elementsToAdd = (Array.isArray(arrayElement) ? arrayElement : [arrayElement]).filter(el => elementArray.indexOf(el) === -1);
    if (elementsToAdd.length === 0) {
      return doc;
    }
    const insertIndex = insertIntoIndex < 0 || insertIntoIndex >= elementArray.length ? elementArray.length : insertIntoIndex;
    const updatedArray = elementArray.slice();
    updatedArray.splice(insertIndex, 0, ...elementsToAdd);
    const updateChange = {
      __typename: _HTypes.HDocCommandType.CHANGE_ELEMENT,
      element: {
        __typename: elementType,
        _id: elementId
      },
      // @ts-expect-error
      changes: {
        [arrayFieldName]: updatedArray
      }
    };
    return docReducer(doc, updateChange);
  }
  return doc;
};
exports.addElementToArrayReducer = addElementToArrayReducer;
function idAndTypeOfChange(change, doc) {
  if (change.__typename === _HTypes.HDocCommandType.INSERT_ELEMENT) {
    return (0, _HUtils.isElementId)(change.element) ? change.element : {
      __typename: change.element.__typename,
      _id: 'NOTVALID'
    };
  } else if (change.__typename === _HTypes.HDocCommandType.CHANGE_ELEMENT || change.__typename === _HTypes.HDocCommandType.DELETE_ELEMENT || change.__typename === _HTypes.HDocCommandType.MOVE_ELEMENT) {
    return (0, _HUtils.isElementId)(change.element) ? change.element : doc ? idAndTypeForPath(doc, change.element) : {
      __typename: 'Invalid',
      _id: 'Invalid'
    };
  } else {
    return {
      __typename: 'Invalid',
      _id: 'Invalid'
    };
  }
}
//# sourceMappingURL=HDocument.js.map